
use imgui_glow_renderer::AutoRenderer;
use imgui_sdl2_support::SdlPlatform;
use std::collections::HashMap;
use sdl2::event::Event;
const MODELS: [&str; 3]= ["resources/models/TestCarModel/CarW4.obj", "resources/models/TestCarModel/CarW1.obj",
"resources/models/TestCarModel/CarW0.obj"];

mod info;

pub enum dEvent
    {
        loadModel(i32),
        getObj(i32),
        addWindow(OpenWindows),
        removeWindow(OpenWindows),
        setObj(cgmath::Vector3<f32>),
        exit,
    }


pub enum Callback
    {
        ModelInfo(info::ModelI),
        ObjectInfo(info::ObjectI),
        MainMenu,
    }

#[derive(Clone, PartialEq)]
pub enum OpenWindows {
    ModelInfo,
    MainMenu,
    Object(i32),
}

pub struct Debug {
    imgui: imgui::Context,
    renderer: AutoRenderer,
    platform: SdlPlatform,
    pub windows: HashMap<u16, OpenWindows>,
    next_id: u16,
    free_ids:Vec<u16>
}


fn makeMint<T>(input: cgmath::Vector3<T>) -> mint::Vector3<T>
{
    mint::Vector3{
        x: input.x,
        y: input.y,
        z: input.z
    }

}

fn makeCG<T>(input: mint::Vector3<T>) -> cgmath::Vector3<T>
    {
        cgmath::Vector3
            {
                x: input.x,
                y: input.y,
                z: input.z
            }
    }

impl Debug {
    pub fn new(gl: std::sync::Arc<glow::Context>) -> Debug
        {
            let mut imgui = imgui::Context::create();

            imgui
                .fonts()
                .add_font(&[imgui::FontSource::DefaultFontData {config: None}]);

            let mut platform = SdlPlatform::init(&mut imgui);
            let mut renderer = match AutoRenderer::initialize(gl.clone(), &mut imgui)
                {
                    Ok(ret) => ret,
                    Err(s) => panic!("Failed to initialize Imgui Platform Renderer: {}", s),
                };

            let mut windows = HashMap::new();
            windows.insert(0,OpenWindows::MainMenu,);
            Debug {
                imgui,
                renderer,
                platform,
                windows,
                next_id: 1,
                free_ids: Vec::new(),
            }
        }

    pub fn handle(&mut self, event: &sdl2::event::Event)
        {
            self.platform.handle_event(&mut self.imgui, event);
        }

    pub fn drawImgui(&mut self, events_loop: &sdl2::EventPump, window: &sdl2::video::Window,
    callback: Vec<Callback>) -> Vec<dEvent>
        {
            self.platform.prepare_frame(&mut self.imgui, &window, &events_loop);
            let ui = self.imgui.new_frame();
            let mut ret = Vec::new();
            for call in callback
            {
                match call {
                    Callback::ModelInfo(a) =>
                        ret.append(&mut Debug::displayModel(ui, &a)),
                    Callback::MainMenu =>
                        ret.append(&mut Debug::displayMain(ui)),
                    Callback::ObjectInfo(mut a) =>
                        ret.append(&mut Debug::displayObject(ui, &mut a))
                }

            }



            let draw_data = self.imgui.render();

            self.renderer.render(draw_data).unwrap();

            ret
        }
    pub fn addWindow(&mut self, window: OpenWindows)
    {
        let mut found = false;
        for (hashId, windowType) in &self.windows
            {
                if *windowType == window
                    {
                        found = true;
                        break;
                    }
            }
        if (!found)
            {
        if (self.free_ids.is_empty())
            {
                self.windows.insert(self.next_id, window);
                self.next_id += 1;
            }
        else {
            self.windows.insert(self.free_ids.pop().unwrap(), window);
        }
            }
    }
    pub fn removeWindow(&mut self, window: OpenWindows)
        {
            println!("searching for a match...");
            let mut id = 0;
            for (hashId, windowType) in &self.windows
            {
                if *windowType == window {
                    id = *hashId;
                    break;
                }

            }
            if id !=0
            {
                self.windows.remove(&id);
                self.free_ids.push(id);
            }
        }


    fn displayModel(ui: &mut imgui::Ui ,model: &info::ModelI) -> Vec<dEvent>
    {
        let mut ret = Vec::new();
        ui.window("Stats")
          .always_auto_resize(true)
          .build(|| {
              ui.text("Model Information");
              ui.text(format!("count meshes: {}", model.meshesCount));
              ui.separator();
              for i in 0..model.meshesCount
              {
                  ui.text(format!("Mesh Number {}", i));
                  ui.text(format!("count Vertices: {}", model.vertCount[i]));
                  ui.separator();
              }
              if ui.button("close") {
                  ret.push(dEvent::removeWindow(OpenWindows::ModelInfo))
              }
          });
        ret
    }

    fn displayMain(ui: &mut imgui::Ui) -> Vec<dEvent>
    {
        let mut ret = Vec::new();
        ui.window("Welcome to the main Menu")
          .always_auto_resize(true)
          .build(|| {
              ui.text("this is the main Debugger Menu, it's used to Debug the game");
              ui.text("although it isn't done yet");
              ui.separator();
              if ui.button("Open the Model Info") {
                  ret.push(dEvent::addWindow(OpenWindows::ModelInfo))
              }
              if ui.button("Open Object Modifier") {
                  ret.push(dEvent::addWindow(OpenWindows::Object(0)))
              }
              if ui.button("Quit") {
                  ret.push(dEvent::exit)
              }

          });
        ret
    }
    fn displayObject(ui: &mut imgui::Ui , object: &mut info::ObjectI ) -> Vec<dEvent>
    {
        let mut ret = Vec::new();
        let mut position = makeMint(object.Position);
        ui.window("ObjectInfo")
          .always_auto_resize(true)
            .build(|| {
                ui.text("Position");
                ui.input_float3("Position", &mut position).build();
                ui.text("WIP");

                ret.push(dEvent::setObj(makeCG(position)));
            });
        ret

    }
}


use crate::model;
pub fn ModelInfo(model: &model::Model) -> Callback
    {
        let mut vertCount = Vec::new();
        let meshesCount = model.meshes.len();
        for i in 0..model.meshes.len() {
            vertCount.push(model.meshes[i].vertices.len());
        }
        Callback::ModelInfo(info::ModelI {
            meshesCount,
            vertCount
        })
    }

use crate::scene::objects::Player;

pub fn ObjectInfo(object: &Player) -> Callback
    {
        let Position = object.Transform.Position;
        let Rotation = object.Transform.Rotation;
        let Velocity = object.Transform.Velocity;
        let Scale = object.Transform.Scale;
        Callback::ObjectInfo(info::ObjectI {
            Position,
            Rotation,
            Velocity,
            Scale
        })
    }
